
// Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
// All rights reserved.  Use of this source code is governed by
// a BSD-style license that can be found in the LICENSE file.

#ifndef TOAST_INTERNAL_HPP
#define TOAST_INTERNAL_HPP

#include <toast.hpp>

// This header declares code whose member variables and/or function signatures
// contain types that are conditionally compiled.

#include <toast/atm.hpp>
#include <toast/math_fft_fftw.hpp>
#include <toast/math_fft_mkl.hpp>

#endif // ifndef TOAST_INTERNAL_HPP
