
// Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
// All rights reserved.  Use of this source code is governed by
// a BSD-style license that can be found in the LICENSE file.

#ifndef TOAST_TEST_HPP
#define TOAST_TEST_HPP


namespace toast {
int test_runner(int argc, char * argv[]);
}

#endif // ifndef TOAST_TEST_HPP
