
// Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
// All rights reserved.  Use of this source code is governed by
// a BSD-style license that can be found in the LICENSE file.

#include <toast.hpp>


int main(int argc, char * argv[]) {
    return toast::test_runner(argc, argv);
}
