.. _utilsvmath:

Vector Math Operations
-----------------------------------

The following functions are just simple wrappers vendor math libraries or the default
libm.  They exist mainly just to provide a common interface to architecture-specific
math libraries.

.. autofunction:: toast.utils.vsin

.. autofunction:: toast.utils.vcos

.. autofunction:: toast.utils.vsincos

.. autofunction:: toast.utils.vatan2

.. autofunction:: toast.utils.vsqrt

.. autofunction:: toast.utils.vrsqrt

.. autofunction:: toast.utils.vexp

.. autofunction:: toast.utils.vlog
